#include "queue.h"                            // ns-2
#include "delay.h"                            // ns-2
#include "agent.h"                            // ns-2

#include <list>                               // list
#include <map>                                // multimap

using namespace std;

class InstVar {
public:
  static double bw_atof(const char* s);
  static double time_atof(const char* s);
};

namespace DSF {

class DSFQueue;
typedef int64_t PacketTime;
#define DSFPACKET Packet*
#define DSFQ DSFQueue*
#include "dsfbase.h"

class DSFQueue : public Queue {
  // configuration
  DSFBase base;
  int linkRate;
  int queueLimit;
  bool fifo;

  // tracing
  Tcl_Channel tchan_;

public:
  DSFQueue(bool f) : linkRate(0), queueLimit(0), fifo(f), tchan_(0) {}

  virtual void reset() {
    if (fifo) base.setQueueLimit(queueLimit);
    base.reset();
  }

  virtual void enque( Packet* p ) {
    int cnum = HDR_IP(p)->prio() % DSFBase::codeCount;
    base.enqueue(p, cnum, this);
  }

  virtual Packet* deque() {
    double delay = 0;
    return base.dequeue( delay, this );
  }

  static void Drop( DSFQueue* This, Packet* p );
  static void Trace( DSFQueue* This, char* buf );

  virtual int command( int argc, const char*const* argv );
};

#define DSFPLEN(p) HDR_CMN(p)->size()
#define DSFNOW Scheduler::instance().clock()
#define DSFASSERT(cond) assert(cond)
#define DSFDROP(caller, p) DSFQueue::Drop(caller, p)
#define DSF:DROP(caller, p) DSFQueue::Drop(caller, p)
#define DSFTRACE(caller, b) DSFQueue::Trace(caller, b)
#include "dsfbase2.h"

int DSFQueue::command( int argc, const char*const* argv ) {
  Tcl& tcl = Tcl::instance();
  switch (argc) {
  case 2:
    if (strcmp(argv[1], "set-abe") == 0) {
      base.setABE();
      return (TCL_OK);
    }
    break;
  case 3:
    if (strcmp(argv[1], "attach") == 0) {
      int mode;
      tchan_ = Tcl_GetChannel( tcl.interp(), (char*)argv[2], &mode );
      if ( !tchan_ ) {
        tcl.resultf( "DSF: can't attach trace %s for writing", argv[2] );
        base.traceflag = false;
        return (TCL_ERROR);
      }
      base.traceflag = true;
      return (TCL_OK);
    }
    if (strcmp(argv[1], "target") == 0) {
      LinkDelay* del = (LinkDelay*)TclObject::lookup( argv[2] );
      if ( del == 0 ) {
        tcl.resultf( "DSF: no LinkDelay object %s", argv[2] );
        return (TCL_ERROR);
      }
      // second "target" call comes from flow monitor attachment (BW 0)
      if ( !linkRate ) linkRate = del->bandwidth() / 8;
      return Queue::command( argc, argv );
    }
    if (strcmp(argv[1], "set-queue-limit") == 0) {
      if ( !linkRate ) {
        tcl.resultf( "DSF: link rate not set" );
        return (TCL_ERROR);
      }
      queueLimit = linkRate * InstVar::time_atof(argv[2]);
      fifo = true;
//      fprintf( stderr, "DSF: queueLimit %d\n", queueLimit);
      return (TCL_OK);
    }
    break;
  case 4:
    if (strcmp(argv[1], "add-class") == 0) {
      int cnum = atoi(argv[2]);
      if ( cnum >= DSFBase::codeCount ) {
        tcl.resultf( "DSF: illegal code point %d > %d", cnum, DSFBase::codeCount );
        return (TCL_ERROR);
      }
      PacketTime td = PacketTime(linkRate * InstVar::time_atof(argv[3]));
//      fprintf( stderr, "DSF: add code point %d with delay %ld bytes\n", cnum, td );
      base.addClass( cnum, td );
      return (TCL_OK);
    }
    break;
  }
  return Queue::command(argc, argv);
}

void DSFQueue::Trace( DSFQueue* This, char* buf ) {
  Tcl_Write( This->tchan_, buf, strlen(buf) );
  Tcl_Write( This->tchan_, "\n", 1);
}

void DSFQueue::Drop( DSFQueue* This, Packet* p ) {
  This->drop(p);
}

// hook into Tcl for object creation
static class DSFClass : public TclClass {
public:
  DSFClass() : TclClass("Queue/DSF") {}
  TclObject* create(int, const char*const*) { return new DSFQueue(false); }
} class_dsf;

// hook into Tcl for object creation
static class FIFOClass : public TclClass {
public:
  FIFOClass() : TclClass("Queue/FIFO") {}
  TclObject* create(int, const char*const*) { return new DSFQueue(true); }
} class_fifo;

};
