/*
 * a and b initiates a bulk send TCP transfer to c.
 * additional isochronous traffic (VoIP-like) between a-c, b-c
 * The default TCP version in ns-3, TcpNewReno, is used as the transport-layer
 * protocol.
 */

#include <iostream>
#include <fstream>
#include <string>

#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/internet-module.h"
#include "ns3/point-to-point-module.h"        // PointToPointHelper
#include "ns3/point-to-point-layout-module.h" // PointToPointDumbbellHelper
#include "ns3/applications-module.h"
#include "ns3/config-store-module.h"
#include "ns3/error-model.h"
#include "ns3/tcp-header.h"
#include "ns3/udp-header.h"
#include "ns3/enum.h"
#include "ns3/event-id.h"
#include "ns3/ipv4-global-routing-helper.h"
#include "ns3/ipv4-flow-classifier.h"
#include "ns3/flow-monitor-helper.h"

#define DSFDCE 1

#if DSFDCE
#include "ns3/dce-module.h"
#include "ns3/linux-stack-helper.h"
#endif

using namespace ns3;

NS_LOG_COMPONENT_DEFINE ("CoDelTest");

static const uint64_t mega = 1000000;
std::string desc;
uint64_t run;
static bool linuxStack = false;

void
printStats (FlowMonitor::FlowStats st, FlowId id, Ipv4Address sourceIP, uint16_t sourcePort, Ipv4Address destIP, uint16_t destPort, float simDuration)
{
  std::cout << desc << "," << run << "," << id << ",sIP," << sourceIP << std::endl;
  std::cout << desc << "," << run << "," << id << ",sPo," << sourcePort << std::endl;
  std::cout << desc << "," << run << "," << id << ",dIP," << destIP << std::endl;
  std::cout << desc << "," << run << "," << id << ",dPo," << destPort << std::endl;
  std::cout << desc << "," << run << "," << id << ",Dur," << simDuration << std::endl;
  std::cout << desc << "," << run << "," << id << ",TxB," << st.txBytes << std::endl;
  std::cout << desc << "," << run << "," << id << ",RxB," << st.rxBytes << std::endl;
  std::cout << desc << "," << run << "," << id << ",TxP," << st.txPackets << std::endl;
  std::cout << desc << "," << run << "," << id << ",RxP," << st.rxPackets << std::endl;
  std::cout << desc << "," << run << "," << id << ",LsP," << st.lostPackets << std::endl;
  if (st.rxPackets > 0)
    {
      std::cout << desc << "," << run << "," << id << ",MDl," << (st.delaySum.GetSeconds() / st.rxPackets) << std::endl;
    }
  for (uint32_t i=0; i<st.delayHistogram.GetNBins (); i++)
    std::cout << desc << "," << run << "," << id << "," << st.delayHistogram.GetBinStart (i) << "-"<< st.delayHistogram.GetBinEnd (i) << "," <<st.delayHistogram.GetBinCount (i) << std::endl;
} 

ApplicationContainer
CreateTcpSink (uint16_t port, Ptr< Node > sink, float stopTime)
{
  std::string proto;
#if DSFDCE
  if (linuxStack) proto = "ns3::LinuxTcpSocketFactory";
  else
#endif
    proto = "ns3::TcpSocketFactory";

  PacketSinkHelper sinkHelper (proto, InetSocketAddress (Ipv4Address::GetAny (), port));
  ApplicationContainer sinkApp = sinkHelper.Install (sink);
#if DSFDCE
  if (linuxStack) sinkApp.Start (Seconds (0.000001));
  else
#endif
    sinkApp.Start (Seconds (0));
  sinkApp.Stop (Seconds (stopTime));
  return sinkApp;
}

ApplicationContainer
CreateUdpSink (uint16_t port, Ptr< Node > sink, float stopTime)
{
  std::string proto;
#if DSFDCE
  if (linuxStack) proto = "ns3::LinuxUdpSocketFactory";
  else
#endif
    proto = "ns3::UdpSocketFactory";

  PacketSinkHelper sinkHelper (proto, InetSocketAddress (Ipv4Address::GetAny (), port));
  ApplicationContainer sinkApp = sinkHelper.Install (sink);
#if DSFDCE
  if (linuxStack) sinkApp.Start (Seconds (0.000001) );
  else
#endif
    sinkApp.Start (Seconds (0));
  sinkApp.Stop (Seconds (stopTime));
  return sinkApp;
}

class GoodputApp {
  static const double waitTime=0.5;
  uint64_t fileSize;
  Ptr<ParetoRandomVariable> size;
  double startTime;
  AddressValue remoteAddress;
  Ptr<Node> sender;
public:
  GoodputApp(double sT, AddressValue rA, Ptr<Node> sdr) :
  startTime(sT+0.000001), remoteAddress(rA), sender(sdr) {
    size = CreateObject<ParetoRandomVariable>();
  }
  void Start() {
    Ptr<BulkSendApplication> app = CreateObject<BulkSendApplication>();
    fileSize = size->GetValue(1*mega, 1.4, 100.0*mega);
    app->SetAttribute ("Remote", remoteAddress);
    std::string proto;
#if DSFDCE
    if (linuxStack) proto = "ns3::LinuxTcpSocketFactory";
    else
#endif
      proto = "ns3::TcpSocketFactory";
    app->SetAttribute( "Protocol", StringValue(proto) );
    app->SetAttribute( "MaxBytes", UintegerValue(fileSize) );
    app->SetAttribute ("SendSize", UintegerValue (1400));
#if DSFDCE
    if (linuxStack && startTime == 0) {
      std::cerr << "cannot start applications at time 0.0" << std::endl;
      exit(1);
    }
#endif
    app->SetAttribute( "StartTime", TimeValue(Seconds(startTime)) );
    app->setDoneCallback(doneCallback, this);
    sender->AddApplication(app);
    startTime += Now().GetSeconds();
  }

  static void doneCallback(void* This) {
    GoodputApp* app = (GoodputApp*)This;
    std::cout << desc << "," << run << "," << app->fileSize << ",CMP," << Now().GetSeconds() - app->startTime << std::endl;
    Simulator::Schedule( Seconds(waitTime), &GoodputApp::Start, app );
    app->startTime = Now().GetSeconds() + waitTime;
  }
};

void
CreateVoipFlow (AddressValue remoteAddress, Ptr<Node> sender, float stopTime, float startTime)
{
  startTime += 0.000001;
  Ptr<OnOffApplication> app = CreateObject<OnOffApplication>();
  app->SetAttribute ("Remote", remoteAddress);
  std::string proto;
#if DSFDCE
  if (linuxStack) proto = "ns3::LinuxUdpSocketFactory";
  else
#endif
     proto = "ns3::UdpSocketFactory";
  app->SetAttribute( "Protocol", StringValue(proto) );
  app->SetAttribute( "PacketSize", UintegerValue(300) );
  app->SetAttribute( "OnTime",StringValue("ns3::ExponentialRandomVariable[Mean=.2]"));
  app->SetAttribute( "OffTime",StringValue("ns3::ExponentialRandomVariable[Mean=1]"));
#if DSFDCE
    if (linuxStack && startTime == 0) {
      std::cerr << "cannot start applications at time 0.0" << std::endl;
      exit(1);
    }
#endif
  app->SetAttribute( "StartTime", TimeValue(Seconds(startTime)) );
  app->SetAttribute( "StopTime", TimeValue(Seconds(stopTime - 3)) );
  app->SetAttribute( "DataRate", DataRateValue (DataRate ("100kb/s")));
  sender->AddApplication(app);
}

void
CreateBgFlow (AddressValue remoteAddress, Ptr<Node> sender, float stopTime)
{
  Ptr<OnOffApplication> app = CreateObject<OnOffApplication>();
  app->SetAttribute ("Remote", remoteAddress);
  std::string proto;
#if DSFDCE
  if (linuxStack) proto = "ns3::LinuxUdpSocketFactory";
  else
#endif
     proto = "ns3::UdpSocketFactory";
  app->SetAttribute( "Protocol", StringValue(proto));
  app->SetAttribute( "PacketSize", UintegerValue(1000) );
  app->SetAttribute( "OnTime", StringValue("ns3::ParetoRandomVariable[Shape=1.6|Mean=0.5]"));
  app->SetAttribute( "OffTime",StringValue("ns3::ParetoRandomVariable[Shape=1.6|Mean=2]"));
  app->SetAttribute( "StartTime", TimeValue(Seconds(0.1)) );
  app->SetAttribute( "StopTime", TimeValue(Seconds(stopTime)) );
  app->SetAttribute( "DataRate", DataRateValue (DataRate ("850kb/s")));
  sender->AddApplication(app);
}

int main (int argc, char *argv[])
{
  std::string hostDelay = "10ms";
  std::string routerDelay = "10ms";
  std::string queueType = "CoDel";
  double bottleneckRate = 25;
  double defaultRate = 1000;

  uint64_t queueTime   = 200;        // queue time in ms
  uint32_t numOfBulkFlows = 20;   // # of download bulk transfer flows
  uint32_t numOfVoipFlows = 10;   // # of download onoff flows
  uint32_t numOfBgFlows = 30;   // # of background pareto flows

  float waitStable = 5;

  run = 1;
  desc = queueType;

  float simDuration = 60;        //in seconds
  double delayBinWidth=0.001;
  float codelInterval = 100;      // The CoDel algorithm interval in ms
  float codelTarget = 5;          // The CoDel algorithm target queue delay in ms

  CommandLine cmd;
  cmd.AddValue ("queueType", "bottleneck queue type", queueType);
  cmd.AddValue ("queueTime",   "queue time in ms",    queueTime);
  cmd.AddValue ("numOfBulkFlows", "number of bulk transfer flows", numOfBulkFlows);
  cmd.AddValue ("numOfOnOffFlows", "number of voip flows", numOfVoipFlows);
  cmd.AddValue ("numOfBgFlows", "number of bg flows", numOfBgFlows);
  cmd.AddValue ("hostDelay", "delay in ms of each host's link", hostDelay);
  cmd.AddValue ("routerDelay", "delay in ms of the bottleneck link", routerDelay);
  cmd.AddValue ("codelInterval", "the CoDel algorithm interval in ms", codelInterval);
  cmd.AddValue ("codelTarget", "the CoDel algorithm target queue delay in ms", codelTarget);
  cmd.AddValue ("delayBinWidth", "width of delay width in report", delayBinWidth);
  cmd.AddValue ("simDuration", "simulation duration in seconds", simDuration);
  cmd.AddValue ("run", "run number", run);
  cmd.AddValue ("desc", "description of run for output csv", desc);
  cmd.AddValue("linuxStack",  "use linux stack (0)",            linuxStack);
  cmd.Parse (argc, argv);

  bottleneckRate = bottleneckRate * mega;
  defaultRate = defaultRate * mega;
  RngSeedManager::SetRun(run);

  // set up topology
  PointToPointHelper bottleneck;
  bottleneck.SetDeviceAttribute ("DataRate", DataRateValue(DataRate(bottleneckRate)));
  bottleneck.SetChannelAttribute("Delay", StringValue(routerDelay));

  PointToPointHelper leaf;
  leaf.SetDeviceAttribute ("DataRate", DataRateValue(DataRate(defaultRate)));
  leaf.SetChannelAttribute("Delay", StringValue(hostDelay));

  unsigned int nodeCount = numOfBulkFlows + numOfVoipFlows + numOfBgFlows;
  static PointToPointDumbbellHelper* topology = new PointToPointDumbbellHelper(nodeCount, leaf, nodeCount, leaf, bottleneck);

  NodeContainer edges, routers;
  routers.Add(topology->GetLeft());
  routers.Add(topology->GetRight());
  for (unsigned int i = 0; i < nodeCount; i += 1) {
    edges.Add(topology->GetLeft(i));
    edges.Add(topology->GetRight(i));
  }

  InternetStackHelper stack;
#if DSFDCE
  DceManagerHelper dceManager;
  LinuxStackHelper lstack;
  if (linuxStack) {
    GlobalValue::Bind("ChecksumEnabled", BooleanValue(true));
    dceManager.SetNetworkStack("ns3::LinuxSocketFdFactory", "Library", StringValue ("liblinux.so"));
    dceManager.Install(edges);
    lstack.Install(edges);
    lstack.SysctlSet(edges, ".net.ipv4.tcp_rmem", "4096	87380	375000");
    lstack.SysctlSet(edges, ".net.ipv4.tcp_wmem", "4096	16384	375000");
    lstack.SysctlSet(edges, ".net.ipv4.tcp_allowed_congestion_control", "reno cubic");
    lstack.SysctlSet(edges, ".net.ipv4.tcp_congestion_control", "cubic");
  } else
#endif
    stack.Install(edges);

  stack.Install(routers);

  topology->AssignIpv4Addresses( Ipv4AddressHelper ("10.1.1.0", "255.255.255.0"),
                                 Ipv4AddressHelper ("10.2.1.0", "255.255.255.0"),
                                 Ipv4AddressHelper ("10.3.1.0", "255.255.255.0") );
  Ipv4GlobalRoutingHelper::PopulateRoutingTables();

#if DSFDCE
  if (linuxStack) {
    // setup static routing tables
    LinuxStackHelper::PopulateRoutingTables();
  }
  DceApplicationHelper process;
#endif

  Ptr<PointToPointNetDevice> dev;
  for (int i=0; i<2; i++)
    {
      dev = topology->GetLeft()->GetDevice(i)->GetObject<PointToPointNetDevice>();

      if (queueType.compare ("DropTail") == 0)
	{
	  Ptr<DropTailQueue> queue = CreateObject<DropTailQueue> ();
	  queue->SetAttribute ("Mode", StringValue ("QUEUE_MODE_BYTES"));
	  queue->SetAttribute ("MaxBytes", UintegerValue ( (double(queueTime) / 8000) * bottleneckRate));
	  dev->SetQueue(queue);
	}
      else if (queueType.compare ("CoDel") == 0)
	{
	  Ptr<CoDelQueue> queue = CreateObject<CoDelQueue> ();
	  queue->SetAttribute ("Mode", StringValue ("QUEUE_MODE_BYTES"));
	  queue->SetAttribute ("MaxBytes", UintegerValue ( (double(queueTime) / 8000) * bottleneckRate));
	  queue->SetAttribute ("Interval", TimeValue(MilliSeconds(codelInterval)));
	  queue->SetAttribute ("Target", TimeValue(MilliSeconds(codelTarget)));
	  dev->SetQueue(queue);
	}
    }

  ApplicationContainer sinkApps;
  AddressValue remoteAddress;
  uint32_t ni = 0;
  Ptr<UniformRandomVariable> rstartTime = CreateObject<UniformRandomVariable> ();

  uint16_t port = 40002;
  uint32_t flowi=numOfBulkFlows;
  while (flowi--)
    {
      sinkApps.Add ( CreateTcpSink (port, topology->GetRight(ni), simDuration) );
      remoteAddress = AddressValue (InetSocketAddress (topology->GetRightIpv4Address(ni), port));
      GoodputApp* da = new GoodputApp(float(rstartTime->GetInteger (0,0+20))/10, remoteAddress, topology->GetLeft(ni++));
      da->Start();
    }

  port = 40001;
  flowi=numOfVoipFlows;
  while (flowi--)
    {
      sinkApps.Add (CreateUdpSink (port, topology->GetRight(ni), simDuration));
      remoteAddress = AddressValue (InetSocketAddress(topology->GetRightIpv4Address(ni),port));
      CreateVoipFlow (remoteAddress, topology->GetLeft(ni++), simDuration, float(rstartTime->GetInteger (0,0+20))/10);
    }

  // bg traffic
  port = 49000;
  flowi=numOfBgFlows;
  while (flowi--)
    {
      sinkApps.Add (CreateUdpSink (port, topology->GetRight(ni), simDuration));
      remoteAddress = AddressValue (InetSocketAddress(topology->GetRightIpv4Address(ni),port));
      CreateBgFlow (remoteAddress, topology->GetLeft(ni++), simDuration);
    }


  // Flow Monitor
  FlowMonitorHelper flowmonHelper;
  Ptr<FlowMonitor> monitor; 
  monitor = flowmonHelper.InstallAll ();
  monitor->SetAttribute("DelayBinWidth", DoubleValue(delayBinWidth));
  monitor->Start(Seconds(waitStable));

  Simulator::Stop (Seconds (simDuration));
  Simulator::Run ();

  monitor->CheckForLostPackets ();
  std::map< FlowId, FlowMonitor::FlowStats > stats = monitor->GetFlowStats();
  Ptr<Ipv4FlowClassifier> classifier = DynamicCast<Ipv4FlowClassifier> (flowmonHelper.GetClassifier ());
  for (std::map<FlowId, FlowMonitor::FlowStats>::const_iterator i = stats.begin (); i != stats.end (); ++i)
    {
      Ipv4FlowClassifier::FiveTuple t = classifier->FindFlow (i->first);
      if (t.destinationPort==40001)
	{
	  printStats(i->second, i->first, t.sourceAddress, t.sourcePort, t.destinationAddress, t.destinationPort, simDuration);
	}
    }

  Ptr<PacketSink> pktsink;
  pktsink = sinkApps.Get (0)->GetObject<PacketSink> ();
  std::cout << desc << "," << run << ",tcp,RxS," << pktsink->GetTotalRx () << std::endl;
  pktsink = sinkApps.Get (1)->GetObject<PacketSink> ();
  std::cout << desc << "," << run << ",voip,RxS," << pktsink->GetTotalRx () << std::endl;
  pktsink = sinkApps.Get (2)->GetObject<PacketSink> ();
  std::cout << desc << "," << run << ",bg,RxS," << pktsink->GetTotalRx () << std::endl;


  Simulator::Destroy ();
  return 0;
}
